/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "UploadManager.h"

#include "ConnectionManager.h"
#include "LogManager.h"
#include "ShareManager.h"
#include "ClientManager.h"
#include "FilteredFile.h"
#include "ZUtils.h"
#include "ResourceManager.h"
#include "HashManager.h"
#include "AdcCommand.h"
#include "FavoriteManager.h"

bool UploadManager::m_boFireball = false;
bool UploadManager::m_boFileServer = false;

static const string UPLOAD_AREA = "Uploads";

Upload::Upload(UserConnection& conn) : Transfer(conn), stream(0) { 
	conn.setUpload(this);
}

Upload::~Upload() { 
	getUserConnection().setUpload(NULL);
	delete stream; 
}

void Upload::getParams(const UserConnection& aSource, StringMap& params) {
	Transfer::getParams(aSource, params);
	params["source"] = getSourceFile();
}

UploadManager::UploadManager() throw() : running(0), extra(0), lastGrant(0), mUploadLimit(0), 
	mBytesSent(0), mBytesSpokenFor(0), mCycleTime(0), mByteSlice(0), mThrottleEnable(BOOLSETTING(THROTTLE_ENABLE)), 
	m_boLastTickHighSpeed(false), m_iHighSpeedStartTick(0), boFireballSent(false), boFileServerSent(false) {	
	ClientManager::getInstance()->addListener(this);
	TimerManager::getInstance()->addListener(this);
	throttleZeroCounters();
}

UploadManager::~UploadManager() throw() {
	TimerManager::getInstance()->removeListener(this);
	ClientManager::getInstance()->removeListener(this);

	{
		Lock l(cs);
		for(UploadQueueItem::UserMapIter ii = UploadQueueItems.begin(); ii != UploadQueueItems.end(); ++ii) {
			for(UploadQueueItem::Iter i = ii->second.begin(); i != ii->second.end(); ++i) {
				(*i)->dec();
			}
		}
		UploadQueueItems.clear();
	}

	while(true) {
		{
			Lock l(cs);
			if(uploads.empty())
				break;			
		}
		Thread::sleep(100);
	}
}

bool UploadManager::prepareFile(UserConnection& aSource, const string& aType, const string& aFile, int64_t aStartPos, int64_t aBytes) {
	if(aFile.empty() || aStartPos < 0 || aBytes < -1 || aBytes == 0) {
		aSource.fileNotAvail("Invalid request");
		return false;
	}
	
	InputStream* is = NULL;
	int64_t start = 0;
	int64_t bytesLeft = 0;
	int64_t size = 0;

	bool userlist = (aFile == Transfer::USER_LIST_NAME_BZ || aFile == "MyList.DcLst");
	bool free = userlist;
	bool leaves = false;

	string sourceFile;
	try {
		if(aType == Transfer::TYPE_FILE) {
			sourceFile = ShareManager::getInstance()->toReal(aFile);
			File* f = new File(sourceFile, File::READ, File::OPEN);

			start = aStartPos;
			size = f->getSize();
			bytesLeft = (aBytes == -1) ? size-aStartPos : aBytes;
		
			if(size < (start + bytesLeft)) {
				aSource.fileNotAvail();
				delete f;
				return false;
			}

			free = free || (size <= (int64_t)(SETTING(SMALL_FILE_SIZE) * 1024));

			f->setPos(start);

			is = f;
			if((start + bytesLeft) < size) {
				is = new LimitedInputStream<true>(is, aBytes);
				aSource.unsetFlag(UserConnection::FLAG_END_ON_FILE_END);
			} else {
				aSource.setFlag(UserConnection::FLAG_END_ON_FILE_END);
			}
		} else if(aType == Transfer::TYPE_TTHL) {
			sourceFile = ShareManager::getInstance()->toReal(aFile);
			MemoryInputStream* mis = ShareManager::getInstance()->getTree(aFile);
			if(mis == NULL) {
				aSource.fileNotAvail();
				return false;
			}
			start = 0;
			bytesLeft = size = mis->getSize();
			is = mis;
			leaves = true;
			free = true;
		} else {
			aSource.fileNotAvail("Unknown file type");
			return false;
		}
	} catch(const ShareException& e) {
		aSource.fileNotAvail(e.getError());
		return false;
	} catch(const Exception& e) {
		LogManager::getInstance()->message(STRING(UNABLE_TO_SEND_FILE) + sourceFile + ": " + e.getError());
		aSource.fileNotAvail();
		return false;
	}
	
	if(userlist == false && aSource.isSet(UserConnection::FLAG_NOT_AVAIL) == true) {
		aSource.maxedOut();
		delete is;
		return false;
	}

	Lock l(cs);

	bool extraSlot = false;
		
	if(aSource.isSet(UserConnection::FLAG_HASSLOT) == false) {
		bool hasReserved = (reservedSlots.find(aSource.getUser()) != reservedSlots.end());
		bool isFavorite = FavoriteManager::getInstance()->hasSlot(aSource.getUser());

		if(!(hasReserved || isFavorite || getFreeSlots() > 0 || getAutoSlot())) {
			bool supportsFree = aSource.isSet(UserConnection::FLAG_SUPPORTS_MINISLOTS);
			bool allowedFree = aSource.isSet(UserConnection::FLAG_HASEXTRASLOT) || aSource.isSet(UserConnection::FLAG_OP) || getFreeExtraSlots() > 0;
			if(free && supportsFree && allowedFree) {
				extraSlot = true;
			} else {
				delete is;
				aSource.maxedOut();
				if(BOOLSETTING(NO_SEG_UP) == false || aSource.isSet(UserConnection::FLAG_END_ON_FILE_END) == true)
				    addFailedUpload(aSource.getUser(), sourceFile, aStartPos, File::getSize(sourceFile));
				    
				aSource.disconnect();
				return false;
			}
		} 
        
		setLastGrant(GET_TICK());
	}

	clearUserFiles(aSource.getUser());

    if(BOOLSETTING(NO_SEG_UP) == true && aType == Transfer::TYPE_FILE && 
		aSource.isSet(UserConnection::FLAG_END_ON_FILE_END) == false) {
        delete is;
        aSource.error("Badrequest, not supported!");
        return false;
    }

	Upload* u = new Upload(aSource);
	u->setStream(is);
	if(aBytes == -1)
		u->setSize(size);
	else
		u->setSize(start + bytesLeft);

	u->setStartPos(start);
	u->setSourceFile(sourceFile);

	if(userlist)
		u->setFlag(Upload::FLAG_USER_LIST);
	if(leaves)
		u->setFlag(Upload::FLAG_TTH_LEAVES);

	uploads.push_back(u);
	throttleSetup();
	
	if(aSource.isSet(UserConnection::FLAG_HASSLOT) == false) {
		if(extraSlot) {
			if(aSource.isSet(UserConnection::FLAG_HASEXTRASLOT) == false) {
				aSource.setFlag(UserConnection::FLAG_HASEXTRASLOT);
				extra++;
			}
		} else {
			if(aSource.isSet(UserConnection::FLAG_HASEXTRASLOT)) {
				aSource.unsetFlag(UserConnection::FLAG_HASEXTRASLOT);
				extra--;
			}
			aSource.setFlag(UserConnection::FLAG_HASSLOT);
			running++;
		}
	}
	
	return true;
}

int64_t UploadManager::getRunningAverage() {
	Lock l(cs);
	int64_t avg = 0;
	for(Upload::Iter i = uploads.begin(); i != uploads.end(); ++i) {
		Upload* u = *i;
		avg += (int)u->getRunningAverage();
	}
	return avg;
}

bool UploadManager::getAutoSlot() {
	/** A 0 in settings means disable */
	if(SETTING(MIN_UPLOAD_SPEED) == 0)
		return false;
	/** Only grant one slot per 30 sec */
	if(GET_TICK() < getLastGrant() + 30*1000)
		return false;
	/** Grant if upload speed is less than the threshold speed */
	return getRunningAverage() < (SETTING(MIN_UPLOAD_SPEED)*1024);
}

void UploadManager::removeUpload(Upload* aUpload) {
	Lock l(cs);
	dcassert(find(uploads.begin(), uploads.end(), aUpload) != uploads.end());
	uploads.erase(remove(uploads.begin(), uploads.end(), aUpload), uploads.end());
	throttleSetup();
	delete aUpload;
}

void UploadManager::reserveSlot(const User::Ptr& aUser) {
	{
		Lock l(cs);
		reservedSlots[aUser] = GET_TICK() + 600*1000;
	}
	if(aUser->isOnline())
		ClientManager::getInstance()->connect(aUser, Util::toString(Util::rand()));
}
void UploadManager::reserveSlotHour(const User::Ptr& aUser) {
	{
		Lock l(cs);
		reservedSlots[aUser] = GET_TICK() + 3600*1000;
	}
	if(aUser->isOnline())
		ClientManager::getInstance()->connect(aUser, Util::toString(Util::rand()));
}

void UploadManager::reserveSlotDay(const User::Ptr& aUser) {
	{
		Lock l(cs);
		reservedSlots[aUser] = GET_TICK() + 24*3600*1000;
	}
	if(aUser->isOnline())
		ClientManager::getInstance()->connect(aUser, Util::toString(Util::rand()));
}

void UploadManager::reserveSlotWeek(const User::Ptr& aUser) {
	{
		Lock l(cs);
		reservedSlots[aUser] = GET_TICK() + 7*24*3600*1000;
	}
	if(aUser->isOnline())
		ClientManager::getInstance()->connect(aUser, Util::toString(Util::rand()));
}

void UploadManager::unreserveSlot(const User::Ptr& aUser) {
	SlotIter uis = reservedSlots.find(aUser);
	if(uis != reservedSlots.end())
		reservedSlots.erase(uis);
}

void UploadManager::on(UserConnectionListener::Get, UserConnection* aSource, const string& aFile, int64_t aResume) throw() {
	if(aSource->getState() != UserConnection::STATE_GET) {
		dcdebug("UM::onGet Bad state, ignoring\n");
		return;
	}
	
	if(prepareFile(*aSource, Transfer::TYPE_FILE, Util::toAdcFile(aFile), aResume, -1)) {
		aSource->setState(UserConnection::STATE_SEND);
		aSource->fileLength(Util::toString(aSource->getUpload()->getSize()));
	}
}

void UploadManager::onGetBlock(UserConnection* aSource, const string& aFile, int64_t aStartPos, int64_t aBytes, bool z) {
	if(aSource->getState() != UserConnection::STATE_GET) {
		dcdebug("UM::onGet Bad state, ignoring\n");
		return;
	}
	
	if(!z || BOOLSETTING(COMPRESS_TRANSFERS)) {
		if(prepareFile(*aSource, Transfer::TYPE_FILE, Util::toAdcFile(aFile), aStartPos, aBytes)) {
			Upload* u = aSource->getUpload();
			dcassert(u != NULL);
			if(aBytes == -1)
				aBytes = u->getSize() - aStartPos;

			dcassert(aBytes >= 0);

			u->setStart(GET_TICK());

			if(z) {
				u->setStream(new FilteredInputStream<ZFilter, true>(u->getStream()));
				u->setFlag(Upload::FLAG_ZUPLOAD);
			}

			aSource->sending(aBytes);
			aSource->setState(UserConnection::STATE_RUNNING);
			aSource->transmitFile(u->getStream());
			fire(UploadManagerListener::Starting(), u);
		}
	}
}

void UploadManager::on(UserConnectionListener::Send, UserConnection* aSource) throw() {
	if(aSource->getState() != UserConnection::STATE_SEND) {
		dcdebug("UM::onSend Bad state, ignoring\n");
		return;
	}

	Upload* u = aSource->getUpload();
	dcassert(u != NULL);

	u->setStart(GET_TICK());
	aSource->setState(UserConnection::STATE_RUNNING);
	aSource->transmitFile(u->getStream());
	fire(UploadManagerListener::Starting(), u);
}

void UploadManager::on(AdcCommand::GET, UserConnection* aSource, const AdcCommand& c) throw() {
	int64_t aBytes = Util::toInt64(c.getParam(3));
	int64_t aStartPos = Util::toInt64(c.getParam(2));
	const string& fname = c.getParam(1);
	const string& type = c.getParam(0);

	if(prepareFile(*aSource, type, fname, aStartPos, aBytes)) {
		Upload* u = aSource->getUpload();
		dcassert(u != NULL);

		AdcCommand cmd(AdcCommand::CMD_SND);
		cmd.addParam(type).addParam(fname)
			.addParam(Util::toString(u->getPos()))
			.addParam(Util::toString(u->getSize() - u->getPos()));

		if(c.hasFlag("ZL", 4)) {
			u->setStream(new FilteredInputStream<ZFilter, true>(u->getStream()));
			u->setFlag(Upload::FLAG_ZUPLOAD);
			cmd.addParam("ZL1");
		}

		aSource->send(cmd);

		u->setStart(GET_TICK());
		aSource->setState(UserConnection::STATE_RUNNING);
		aSource->transmitFile(u->getStream());
		fire(UploadManagerListener::Starting(), u);
	}
}

void UploadManager::on(UserConnectionListener::BytesSent, UserConnection* aSource, size_t aBytes, size_t aActual) throw() {
	dcassert(aSource->getState() == UserConnection::STATE_RUNNING);
	Upload* u = aSource->getUpload();
	dcassert(u != NULL);
	u->addPos(aBytes, aActual);
	throttleBytesTransferred(aActual);
}

void UploadManager::on(UserConnectionListener::Failed, UserConnection* aSource, const string& aError) throw() {
	Upload* u = aSource->getUpload();

	if(u) {
		fire(UploadManagerListener::Failed(), u, aError);

		dcdebug("UM::onFailed: Removing upload\n");
		removeUpload(u);
	}

	removeConnection(aSource);
}

void UploadManager::on(UserConnectionListener::TransmitDone, UserConnection* aSource) throw() {
	dcassert(aSource->getState() == UserConnection::STATE_RUNNING);
	Upload* u = aSource->getUpload();
	dcassert(u != NULL);

	aSource->setState(UserConnection::STATE_GET);

	if(BOOLSETTING(LOG_UPLOADS) && (BOOLSETTING(LOG_FILELIST_TRANSFERS) || !u->isSet(Upload::FLAG_USER_LIST)) && 
		!u->isSet(Upload::FLAG_TTH_LEAVES) && (BOOLSETTING(LOG_PARTIAL_UPLOADS) || aSource->isSet(UserConnection::FLAG_END_ON_FILE_END) == true)) {
		StringMap params;
		u->getParams(*aSource, params);
		LOG(LogManager::UPLOAD, params);
	}

	fire(UploadManagerListener::Complete(), u);
	removeUpload(u);
}

void UploadManager::addFailedUpload(User::Ptr& User, string file, int64_t pos, int64_t size) {
	time_t itime = GET_TIME();
	bool found = false;
	UploadQueueItem::UserMapIter j = UploadQueueItems.find(User);
	if(j != UploadQueueItems.end()) {
		for(UploadQueueItem::Iter i = j->second.begin(); i != j->second.end(); ++i) {
			if((*i)->File == file) {
				(*i)->pos = pos;
				found = true;
				break;
			}
		}
	}
	if(found == false) {
		UploadQueueItem* qi = new UploadQueueItem(User, file, pos, size, itime);
		{
			UploadQueueItem::UserMapIter i = UploadQueueItems.find(User);
			if(i == UploadQueueItems.end()) {
				UploadQueueItem::List l;
				l.push_back(qi);
				UploadQueueItems.insert(make_pair(User, l));
			} else {
				i->second.push_back(qi);
			}
		}
		fire(UploadManagerListener::QueueAdd(), qi);
	}
}

void UploadManager::clearUserFiles(const User::Ptr& source) {
	UploadQueueItem::UserMapIter ii = UploadQueueItems.find(source);
	if(ii != UploadQueueItems.end()) {
		for(UploadQueueItem::Iter i = ii->second.begin(); i != ii->second.end(); ++i) {
			fire(UploadManagerListener::QueueItemRemove(), (*i));
			(*i)->dec();
		}
		UploadQueueItems.erase(ii);
		fire(UploadManagerListener::QueueRemove(), source);
	}
}

void UploadManager::removeConnection(UserConnection* aSource) {
	dcassert(aSource->getUpload() == NULL);
	aSource->removeListener(this);
	if(aSource->isSet(UserConnection::FLAG_HASSLOT)) {
		running--;
		aSource->unsetFlag(UserConnection::FLAG_HASSLOT);
	} 
	if(aSource->isSet(UserConnection::FLAG_HASEXTRASLOT)) {
		extra--;
		aSource->unsetFlag(UserConnection::FLAG_HASEXTRASLOT);
	}
}

void UploadManager::on(TimerManagerListener::Minute, uint64_t aTick) throw() {
	User::List disconnects;
	{
		Lock l(cs);
		for(SlotIter j = reservedSlots.begin(); j != reservedSlots.end();) {
			if(j->second < aTick) {
				reservedSlots.erase(j++);
			} else {
				++j;
			}
		}
	
	
	
		if( BOOLSETTING(AUTO_KICK) ) {
			for(Upload::Iter i = uploads.begin(); i != uploads.end(); ++i) {
				Upload* u = *i;
				if(u->getUser()->isOnline()) {
					u->unsetFlag(Upload::FLAG_PENDING_KICK);
					continue;
				}
	
				if(u->isSet(Upload::FLAG_PENDING_KICK)) {
					disconnects.push_back(u->getUser());
					continue;
				}
	
				if(BOOLSETTING(AUTO_KICK_NO_FAVS) && FavoriteManager::getInstance()->isFavoriteUser(u->getUser())) {
					continue;
				}
	
				u->setFlag(Upload::FLAG_PENDING_KICK);
			}
		}
	}

	for(User::Iter i = disconnects.begin(); i != disconnects.end(); ++i) {
		LogManager::getInstance()->message(STRING(DISCONNECTED_USER) + Util::toString(ClientManager::getInstance()->getNicks((*i)->getCID())));
		ConnectionManager::getInstance()->disconnect(*i, false);
	}
}	

void UploadManager::on(GetListLength, UserConnection* conn) throw() {
    if(conn->getUser()->isSet(User::DCPLUSPLUS)) {
        conn->listLen("42");
    } else {
	   conn->listLen(ShareManager::getInstance()->getListLenString());
    }
}

// TimerManagerListener
void UploadManager::on(TimerManagerListener::Second, uint64_t aTick) throw() {
	{
		Lock l(cs);
		
		throttleSetup();
		throttleZeroCounters();
		
		if(uploads.size() > 0)
		  fire(UploadManagerListener::Tick(), uploads);

		fire(UploadManagerListener::QueueUpdate());
	}
	
	if(m_boFireball == false) {
		if(getRunningAverage() >= 102400) {
			if(m_boLastTickHighSpeed) {
				uint64_t iHighSpeedTicks = 0;
				if(aTick >= m_iHighSpeedStartTick) 
					iHighSpeedTicks = (aTick - m_iHighSpeedStartTick);
				else
					iHighSpeedTicks = (aTick + 4294967295 - m_iHighSpeedStartTick);
				
				if(iHighSpeedTicks > 60000) {
					m_boFireball = true;
					if(boFireballSent == false) {
						ClientManager::getInstance()->infoUpdated(true);
						boFireballSent = true;
					}
				}
			} else {
				m_iHighSpeedStartTick = aTick;
				m_boLastTickHighSpeed = true;
			}
		} else {
            m_iHighSpeedStartTick = 0;
			m_boLastTickHighSpeed = false;
		}

		if(m_boFileServer == false) {
			if((Util::getUptime() > 7200) && (ShareManager::getInstance()->getShareSize() > 2147483648) && 
    		  (Socket::getTotalUp() > 209715200)) {
				m_boFileServer = true;
				if(boFireballSent == false && boFileServerSent == false) {
					ClientManager::getInstance()->infoUpdated(true);
					boFileServerSent = true;
				}
			}
		}
	}
}

void UploadManager::on(ClientManagerListener::UserDisconnected, const User::Ptr& aUser) throw() {
	if(!aUser->isOnline()) {
		Lock l(cs);
		clearUserFiles(aUser);
	}
}

size_t UploadManager::throttleGetSlice() {
	if(mThrottleEnable) {
		size_t left = mUploadLimit - mBytesSpokenFor;
		if(left > 0) {
			if(left > 2*mByteSlice) {
				mBytesSpokenFor += mByteSlice;
				return mByteSlice;
			} else {
				mBytesSpokenFor += left;
				return left;
			}
		} else {
			return 16; // must send > 0 bytes or threadSendFile thinks the transfer is complete
		}
	} else {
		return (size_t)-1;
	}
}

size_t UploadManager::throttleCycleTime() {
	if(mThrottleEnable)
		return mCycleTime;
	return 0;
}

void UploadManager::throttleZeroCounters()  {
	if(mThrottleEnable) {
		mBytesSpokenFor = 0;
		mBytesSent = 0;
	}
}

void UploadManager::throttleBytesTransferred(uint32_t i)  {
	mBytesSent += i;
}

void UploadManager::throttleSetup() {
	unsigned int num_transfers = uploads.size();
	mUploadLimit = (SETTING(MAX_UPLOAD_SPEED_LIMIT) * 1024);
	mThrottleEnable = BOOLSETTING(THROTTLE_ENABLE) && (mUploadLimit > 0) && (num_transfers > 0);
	if(mThrottleEnable) {
		size_t inbufsize = SETTING(SOCKET_OUT_BUFFER);
		if(mUploadLimit <= (inbufsize * 10 * num_transfers)) {
			mByteSlice = mUploadLimit / (5 * num_transfers);
			if (mByteSlice > inbufsize)
				mByteSlice = inbufsize;
			mCycleTime = 1000 / 10;
		} else {
			mByteSlice = inbufsize;
			mCycleTime = 1000 * inbufsize / mUploadLimit;
		}
	}
}

UploadQueueItem::UserMap UploadManager::getQueue() {
	return UploadQueueItems;
}

/* Some parts Copyright (C) 2002-2007 Petr Kozelka, PPK at czdc dot org */
